package dev.ninjdai.werewolf.worldgen;

import dev.ninjdai.werewolf.worldgen.noises.BiomeNoises;
import dev.ninjdai.werewolf.worldgen.noises.Noises;
import net.minestom.server.coordinate.Point;
import net.minestom.server.coordinate.Vec;
import net.minestom.server.instance.block.Block;
import net.minestom.server.instance.generator.GenerationUnit;
import net.minestom.server.instance.generator.Generator;
import net.minestom.server.registry.DynamicRegistry;
import net.minestom.server.world.biome.Biome;
import org.jetbrains.annotations.NotNull;

import java.util.Random;

public class LGGenerator implements Generator {
    TerrainBuilder terrainBuilder;
    Random random;
    double densityThreshold = 1;//0.6;
    int upperRelativeDensityLimit = 50;
    int lowerRelativeDensityLimit = 50;

    public LGGenerator(TerrainBuilder tb) {
        this.terrainBuilder = tb;
        this.random = new Random(tb.seed);
    }

    @Override
    public void generate(@NotNull GenerationUnit unit) {
        final Point min = unit.absoluteStart();
        final Point max = unit.absoluteEnd();

        for (int x = min.blockX(); x < max.blockX(); x++) {
            for (int z = min.blockZ(); z < max.blockZ(); z++) {
                generate3D(min, max, x, z, unit);
            }
        }
    }

    private void generate3D(Point min, Point max, int x, int z, GenerationUnit unit) {
        int height = terrainBuilder.getSurfaceHeight(x, z);
        int upperDensityLimit = height + upperRelativeDensityLimit;
        int lowerDensityLimit = height - lowerRelativeDensityLimit;
        int terrainHeight = lowerDensityLimit;

        for (int y = upperDensityLimit; y >= lowerDensityLimit; y--) {
            if (y == lowerDensityLimit) {
                unit.modifier().fill(min.withX(x).withZ(z), new Vec(x + 1, lowerDensityLimit + 1, z + 1), Block.STONE);
                break;
            }

            Vec pos = new Vec(x, y, z);
            if (terrainBuilder.getDensity(x, y, z) > densityThreshold) {
                unit.modifier().setBlock(pos, Block.STONE);
                terrainHeight = Math.max(terrainHeight, y);
            }
        }
        DynamicRegistry.Key<Biome> biome = getBiome(x, terrainHeight, z);
        unit.modifier().setBiome(x, terrainHeight, z, biome);
        decorate(x, terrainHeight, z, unit.fork(unit.absoluteStart(), unit.absoluteEnd()), biome);
    }

    private void decorate(int x, int y, int z, GenerationUnit unit, DynamicRegistry.Key<Biome> biome) {
        double ft = Noises.FEATURES.evaluateNoise(x, z);
        if (biome.equals(Biome.DARK_FOREST)) {
            unit.modifier().setBlock(x, y, z, Block.GRASS_BLOCK);
            for (int i = 1; i < 4; i++) {
                Block b = switch ((int) (3 * Noises.FEATURES.evaluateNoise(x, y - i, z))) {
                    case 0 -> Block.COARSE_DIRT;
                    case 1 -> Block.ROOTED_DIRT;
                    default -> Block.DIRT;
                };
                unit.modifier().setBlock(x, y - i, z, b);
            }
             if (ft > 0.995) {
                 unit.modifier().fill(new Vec(x, y+1, z), new Vec(x+1, y+7, z+1), Block.DARK_OAK_LOG);
            } else if (ft > 0.99) {
                unit.modifier().setBlock(x, y + 1, z, Block.TALL_GRASS);
                unit.modifier().setBlock(x, y + 2, z, Block.TALL_GRASS.withProperty("half", "upper"));
            } else if (ft > 0.8) {
                unit.modifier().setBlock(x, y + 1, z, Block.SHORT_GRASS);
            } else if (ft < 0.01) {
                unit.modifier().setBlock(x, y + 1, z, Block.DANDELION);
            }
            if (!(Noises.ROOFED.evaluateNoise(x, z) > 0.75))
                unit.modifier().fill(new Vec(x, y+6, z), new Vec(x+1, y+9, z+1), Block.DARK_OAK_LEAVES);

        } else if (biome.equals(Biome.DESERT)) {
            unit.modifier().setBlock(x, y, z, Block.SAND);
            for (int i = 1; i < 4; i++) {
                unit.modifier().setBlock(x, y - i, z, Block.SANDSTONE);
            }
            if (ft > 0.995) {
                unit.modifier().setBlock(x, y + 1, z, Block.DEAD_BUSH);
            } else if (ft > 0.99) {
                unit.modifier().setBlock(x, y + 1, z, Block.SHORT_GRASS);
            } else if (ft < 0.01) {
                int height = (int) (1 + 3 * Noises.FEATURES.evaluateNoise(x, y, z));
                for (int i = 0; i < height; i++) {
                    unit.modifier().setBlock(x, y + 1 + i, z, Block.CACTUS);
                }
            }
        } else {
            unit.modifier().setBlock(x, y, z, Block.GRASS_BLOCK);
            for (int i = 1; i < 4; i++) {
                Block b = switch ((int) (2 * Noises.FEATURES.evaluateNoise(x, y - i, z))) {
                    case 0 -> Block.COARSE_DIRT;
                    case 1 -> Block.ROOTED_DIRT;
                    default -> Block.DIRT;
                };
                unit.modifier().setBlock(x, y - i, z, b);
            }
            if (ft > 0.995) {
                int height = (int) (3 + 3 * Noises.FEATURES.evaluateNoise(x, y, z));

                GenerationUnit fork = unit.fork(new Vec(x - 2, y, z - 2), new Vec(x + 3, y + height + 4, z + 3));
                fork.modifier().fill(new Vec(x, y, z), new Vec(x + 1, y + height, z + 1), Block.OAK_LOG);
                fork.modifier().fill(new Vec(x - 1, y + height, z - 2), new Vec(x + 2, y + height + 2, z + 3), Block.OAK_LEAVES);
                fork.modifier().fill(new Vec(x - 2, y + height, z - 1), new Vec(x + 3, y + height + 2, z + 2), Block.OAK_LEAVES);
                fork.modifier().fill(new Vec(x - 1, y + height + 2, z - 1), new Vec(x + 2, y + height + 3, z + 2), Block.OAK_LEAVES);
            } else if (ft > 0.95) {
                unit.modifier().setBlock(x, y + 1, z, Block.TALL_GRASS);
                unit.modifier().setBlock(x, y + 2, z, Block.TALL_GRASS.withProperty("half", "upper"));
            } else if (ft > 0.8) {
                unit.modifier().setBlock(x, y + 1, z, Block.SHORT_GRASS);
            } else if (ft < 0.01) {
                unit.modifier().setBlock(x, y + 1, z, Block.POPPY);
            }
        }
    }

    public DynamicRegistry.Key<Biome> getBiome(int x, int y, int z) {
        double m = BiomeNoises.HUMIDITY.evaluateNoise(x, y, z);
        if (y < 0) return Biome.OCEAN;
        if (y < 10) return Biome.BEACH;

        if (m < 0.16) {
            return Biome.DESERT;
        } else if (m < 0.5) {
            return Biome.PLAINS;
        } else {
            return Biome.DARK_FOREST;
        }
    };
}